import java.io.IOException;
import java.util.ConcurrentModificationException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.json.JSONException;
import org.json.JSONObject;

import se.rupy.http.Daemon;
import se.rupy.http.Event;
import se.rupy.http.Reply;
import se.rupy.http.Service;
import se.rupy.http.Session;

public class Broadcast extends Service {
	public static Log log;
	public static int id;

	public void create(Daemon daemon) {
		log = new Log(daemon);
	}

	public String path() { return "/broadcast"; }

	public void filter(Event event) throws Event, Exception {
		try {
			broadcast(event);
		}
		catch(Exception e) {
			e.printStackTrace();
			throw e;
		}
	}
	
	public void broadcast(Event event) throws Event, Exception {
		event.reply().header("Cache-Control", "no-cache");
		event.query().parse();

		Session session = event.session();
		Channel channel = Poll.room.find(session);

		/*
		 * Disconnect client.
		 */

		if(event.string("type").equals("disconnect")) {
			disconnect(event, session, channel);
			end(event);
		}

		/*
		 * Spam protection.
		 */
		/*
		long time = session.big("time");

		if(System.currentTimeMillis() < time + 1000) {
			session.put("time", System.currentTimeMillis());

			if(!session.bit("spam", false)) {
				session.put("spam", true);
				session.put("send", message(channel, "1b1b1b", "wait 1 second"));
				wakeup(channel.push(session));
			}

			end(event);
		}
		else {
			session.put("spam", false);
		}
		*/
		/*
		 * Commandline tools.
		 */

		String text = event.string("text");

		if(text.equals("/exit") && !channel.name().equals("lobby")) {
			Channel lobby = Poll.room.find("lobby");
			Poll.room.door(session, channel, lobby);
			text = "";
		}

		if(text.equals("/list") || text.equals("/here") || text.equals("/help")) {
			if(text.equals("/list")) {
				session.put("send", message(channel, "1b1b1b", Poll.room.list()));
			}
			else if(text.equals("/here")) {
				session.put("send", message(channel, "1b1b1b", channel.list()));
			}
			else {
				session.put("send", message(channel, "1b1b1b", "/list, /join &lt;name&gt;, /exit, /color &lt;rrggbb&gt;"));
			}

			wakeup(channel.push(session));
			end(event);
		}

		if(text.startsWith("/join") || text.startsWith("/color") || text.startsWith("/host")) {
			int index = text.indexOf(" ");

			if(index > 0) {
				String value = text.substring(index + 1);

				if(text.startsWith("/join") && value.length() > 0 && !channel.name().equals(value)) {
					Channel room = Poll.room.find(value);

					if(room == null) {
						room = Poll.room.add(value, event.bit("xss") ? Poll.XSS_HALLWAY_HOST : null, false);
					}

					Poll.room.door(session, channel, room);
					text = "";
				}
				else if(text.startsWith("/color") && value.length() == 6) {
					session.put("color", clean(value));
					session.put("send", message(channel, value, "new color"));
					wakeup(channel.push(session));
					end(event);
				}
				else if(text.startsWith("/host") && value.length() > 3) {
					// TODO: Verify if host has a talk service running!
					
					JSONObject message = message(channel, "1b1b1b", "host " + value);
					message.put("host", value);
					message.put("channel", channel.name());
					session.put("send", message);

					log.message(session.get("who") + " host -> " + value);

					wakeup(channel.push(session));
					end(event);

					disconnect(event, session, channel);
				}
			}
		}

		text = clean(text);

		/*
		 * Replace http://... with clickable link tag.
		 */

		int index = text.indexOf("http://");

		if(index > -1) {
			int end = text.indexOf(" ", index);
			String link = text.substring(index, end > 0 ? end : text.length());
			text = text.replace(link, "<a href=\"" + link + "\" target=\"_blank\">link</a>");
		}

		/*
		 * Broadcast.
		 */

		if(channel != null && channel.push().size() > 0) {
			session.put("time", System.currentTimeMillis()); // Important so XHR doesent timeout!
			
			if(text.length() > 0) {
				/*
				 * Now broadcasts are filtered on the server too so you can deploy your application module alongside the talk module.
				 */
				
				JSONObject message = new JSONObject((String) event.daemon().send(null, message(channel, session.string("color"), text, session.key()).toString()));
				System.out.println(message);
				if(message == null) {
					end(event);
				}
				else if(message.has("me")) {
					session.put("send", message);
					wakeup(channel.push(session));
					end(event);
				}

				channel.add(message);
				
				log.message(session.get("who") + " -> " + event.string("text"));
			}

			wakeup(channel);
		}

		end(event);
	}

	private void disconnect(Event event, Session session, Channel channel) {
		Event poll = channel.remove(event);

		if(channel != null && poll != null) {
			session.put("length", 0);
			session.put("online", -1);

			Poll.room.remove(channel);

			log.message(session.get("who") + " disconnected");

			wakeup(poll, channel);
		}
	}

	public static String clean(String line) {
		line = line.replaceAll("&", "&#38;"); // leave first
		line = line.replaceAll("\"", "&#34;");
		line = line.replaceAll("'", "&#39;");
		line = line.replaceAll("<", "&#60;");
		line = line.replaceAll(">", "&#62;");
		return line;
	}

	static void end(Event event) throws IOException, Event {
		if(event.bit("xss")) {
			event.reply().type("application/javascript");
			event.output().print("finish_broadcast();");
		}
		else {
			event.output().print("OK");
		}

		throw event;
	}

	static JSONObject message(Channel channel, String color, String text) throws JSONException {
		return message(channel, color, text, "");
	}

	static JSONObject message(Channel channel, String color, String text, String key) throws JSONException {
		JSONObject message = new JSONObject();

		message.put("channel", channel.name());
		message.put("color", color);
		message.put("time", System.currentTimeMillis());
		message.put("text", text);
		message.put("key", key);
		message.put("id", ++id);

		return message;
	}

	static void wakeup(Event event) {
		if(event == null) {
			System.out.println("Poll event was removed.");
			return;
		}
		
		synchronized(event) {
			if(!event.session().bit("lock", false)) {
				event.session().put("lock", true);
				int status = event.reply().wakeup(true);

				if(status == Reply.OK || status == Reply.WORKING) {
					// OK
				}
				else {
					Session session = event.session();
					Channel channel = Poll.room.find(session);
					channel.remove(event);
					System.out.println("User '" + event.session().key() + "' was lost. wakeup()");
				}
			}
		}
	}

	static void wakeup(Channel channel) {
		wakeup(null, channel);
	}

	static void wakeup(Event me, Channel channel) {
		synchronized(channel) {
			Iterator it = channel.push().values().iterator();

			try {
				while(it.hasNext()) {
					Event event = (Event) it.next();

					if(event == null) {
						it.remove();
					}
					else if(event != me) {
						if(event.session().big("time") < System.currentTimeMillis() - Poll.XSS_REFRESH_RATE * 5) {
							it.remove();
							System.out.println("User '" + event.session().key() + "' was lost. wakeup(TIMEOUT)");
						}
						else {
							synchronized(event) {
								if(!event.session().bit("lock", false)) {
									event.session().put("lock", true);
									int status = event.reply().wakeup();

									//System.out.println(status);
									
									// TODO: Here we should add a flag so the next broadcast from 
									// the disconnected user tells him to reconnect his poll socket!

									if(status != Reply.OK) {
										it.remove();
										System.out.println("User '" + event.session().key() + "' was lost. wakeup(COMPLETE)");
									}
								}
							}
						}
					}
				}
			}
			catch(Exception e) {
				it.remove();
				e.printStackTrace();
			}
		}
	}
}
