import java.awt.Color;
import java.io.IOException;
import java.util.Random;

import org.json.JSONArray;
import org.json.JSONObject;

import se.rupy.http.Daemon;
import se.rupy.http.Event;
import se.rupy.http.Service;
import se.rupy.http.Session;

public class Poll extends Service {
	/*
	 * The number of cached messages for each channel
	 * 0.7 > Increased this because custom messages can 
	 * spam the cache and thus push chat messages out.
	 */
	protected static int CHANNEL_CACHE = 25;
	
	/*
	 * How many messages to push back upon join/exit? Has to be less than CHANNEL_CACHE!
	 */
	protected static int THREAD_LENGTH = 10;
	
	/*
	 * How many seconds before empty push to avoid timeout?
	 */
	protected static int XSS_REFRESH_RATE = 60 * 1000;
	
	/*
	 * If you want to run rooms on another server.
	 */
	protected static String XSS_HALLWAY_HOST = null; //System.getProperty("talk", null);
	
	protected static Room room;
	protected static Random random = new Random(System.currentTimeMillis());

	public void create(Daemon daemon) {
		room = new Room();
	}
	
	public void destroy() {
		room.destroy();
	}

	public String path() { return "/poll"; }

	private static String color() {
		return Integer.toHexString(new Color(random.nextInt(256), random.nextInt(256), random.nextInt(256)).getRGB()).substring(2);
	}
	
	public void filter(Event event) throws Event, Exception {
		event.session().put("time", System.currentTimeMillis());

		Session session = event.session();
		Channel channel = room.find(session, event.string("channel"));
		
		//Broadcast.log.message("" + System.identityHashCode(channel));
		
		try {
			poll(event, session, channel);
		}
		catch(Exception e) {
			channel.remove(event);
			e.printStackTrace();
			throw e;
		}
	}
	
	private void poll(Event event, Session session, Channel channel) throws Event, Exception {
		event.reply().header("Cache-Control", "no-cache");
		//event.reply().header("Access-Control-Allow-Origin", "*");
		// This is cross domain XHR... I'm using old school XSS via <script> injection
		event.query().parse();

		if(event.bit("flush")) {
			if(event.bit("xss")) {
				event.reply().type("application/javascript");
				event.output().print("handle_xss_reply();");
			}
			else {
				event.reply().code("504 Timeout");
			}
			
			event.query().put("flush", false);
			event.output().finish();
			throw event;
		}

		boolean override_send = false;
		
		if(!event.push()) {
			Event old = channel.add(event);
			//Broadcast.log.message(old + " " + channel);
			if(old == null) {
				if(!session.containsKey("who")) {
					session.put("who", event.remote());
				}

				if(!session.containsKey("color")) {
					session.put("color", color());
				}
				
				//event.reply().header("P3P", "CP=\"IDC DSP COR ADM DEVi TAIi PSA PSD IVAi IVDi CONi HIS OUR IND CNT\"");

				event.reply().header("P3P", "CP=\"NOI DSP COR NID ADM DEV PSA OUR IND UNI PUR COM NAV INT STA\""); // Facebook...
				
				Broadcast.log.message(session.get("who") + " connected " + event.session().key());
			}
			/*else if(event.query().header("user-agent").indexOf("iPhone") > 0) {
				Broadcast.log.message(event.query().header("user-agent"));
				
				event.session().put("spam", "1");
				
				if(old.session().get("spam") != null)
					override_send = true;
			}*/
			
			/*
			 * TODO: Avoid browser lockup when 
			 * both sockets are long-polling.
			 * 
			int version = event.medium("version");
			
			System.out.print(event + " " + old);
			if(version > 6 && old != null && event.index() != old.index()) {
				try {
					old.query().put("exit", true);
					old.reply().wakeup();
				}
				catch(IOException e) {
					System.out.print(" eeek!");
				}
			}
			System.out.println("");
			 */
		}
		
		//Channel channel = room.find(session);
		
		JSONObject reply = new JSONObject();
		JSONArray array = new JSONArray();
		
		boolean send = false;
		
		if(event.query().bit("page", false)) {
			reply.put("page", true);
			session.put("length", 0);
			send = true;
		}
		
		int index = 0;
		
		if(session.containsKey("first")) {
			array.put(index++, session.get("first"));
			session.remove("first");
			send = true;
		}
		
		/*
		 * This part is the trick, if anyone has a suggestion 
		 * to how this can be done better please tell me!
		 */
		int messages = channel.length() - session.medium("length", 0);
		
		if(messages > 0) {
			int stop = channel.size();
			int start = stop - messages;
			
			start = start < 0 ? 0 : start;

			for(int i = start; i < stop; i++) {
				array.put(index++, channel.get(i));
			}
			
			send = true;
		}

		if(session.containsKey("send")) {
			array.put(index++, session.get("send"));
			session.remove("send");
			send = true;
		}
		
		if(event.bit("exit")) {
			array.put(index++, Broadcast.message(channel, "1b1b1b", "disconnected"));
			send = true;
		}
		
		int users = channel.push().size();
		
		if(users != session.medium("online", -1)) {
			session.put("online", users);

			if(!event.push()) {
				Broadcast.wakeup(event, channel);
			}
			
			send = true;
		}

		if(send && !override_send) {
			reply.put("message", array);
			reply.put("online", users);

			if(event.bit("exit")) {
				reply.put("exit", true);
			}
			
			session.put("length", channel.length());
			
			if(event.bit("xss")) {
				reply.put("cookie", session.key());
				event.reply().type("application/javascript");
				event.output().print("handle_xss_reply(" + reply.toString() + ");");
			}
			else {
				event.output().print(reply.toString());
			}

			event.output().flush();
			event.output().finish();
		}
		else if(!event.push()) {
			synchronized(event) {
				session.put("lock", false);
			}
		}
	}
}
