import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.json.JSONException;

import se.rupy.http.Event;
import se.rupy.http.Reply;
import se.rupy.http.Session;

public class Room implements Runnable {
	private boolean alive;
	private Map room;

	protected Room() {
		room = new HashMap();
		add("lobby", null, true);

		alive = true;
		new Thread(this).start();
	}

	protected void destroy() {
		alive = false;
	}

	public void run() {
		while(alive) {
			try {
				Iterator it = room.values().iterator();

				while(it.hasNext()) {
					Channel channel = (Channel) it.next();

					synchronized(channel) {
						Iterator it2 = channel.push().values().iterator();

						while(it2.hasNext()) {
							Event event = (Event) it2.next();

							if(event.session().big("time") < System.currentTimeMillis() - Poll.XSS_REFRESH_RATE) {
								synchronized(event) {
									//if(event.bit("xss")) {
										event.session().put("lock", true);
										event.query().put("flush", true);
										int status = event.reply().wakeup();

										// TODO: Here we should add a flag so the next broadcast from 
										// the disconnected user tells him to reconnect his poll socket!
										
										if(status != Reply.OK) {
											if(unlock(event, "XSS")) {
												it2.remove();
											}
										}
										/*
									}
									else {
										if(unlock(event, "XHR")) {
											it2.remove();
										}
									}*/
								}
							}
						}
					}

					if(channel.push().size() == 0 && !channel.persistent()) {
						it.remove();
					}
				}
			}
			catch(Exception e) {
				e.printStackTrace();
			}

			try {
				Thread.sleep(1000);
			}
			catch(Exception e) {
				e.printStackTrace();
			}
		}
	}

	private boolean unlock(Event event, String type) {
		if(!event.session().bit("lock", true)) {
			if(type.equals("XSS")) {
				System.out.println("User '" + event.session().key() + "' was lost. flush(" + type + ")");
				return true;
			}
		}
		else {
			event.session().put("time", System.currentTimeMillis() - (Poll.XSS_REFRESH_RATE - 1000));
			event.session().put("lock", false);
		}
		
		return false;
	}
	
	protected synchronized Channel add(String name, String host, boolean persistent) {
		Channel channel = (Channel) room.get(name);

		if(channel == null) {
			channel = new Channel(name, host, persistent);
			room.put(name, channel);
		}
		else {
			return null;
		}

		return channel;
	}

	protected void door(Session session, Channel from, Channel to) throws JSONException {
		Event poll = from.push(session);

		from.remove(poll);
		to.add(poll);

		remove(from);

		session.put("channel", to.name());
		session.put("length", to.length() > Poll.THREAD_LENGTH ? to.length() - Poll.THREAD_LENGTH : 0);
		session.put("first", Broadcast.message(to, "1b1b1b", "joined " + to.name()));

		Broadcast.wakeup(to);
	}

	protected String list() {
		StringBuilder list = new StringBuilder();
		Iterator it = room.keySet().iterator();
		boolean first = true;

		while(it.hasNext()) {
			String name = (String) it.next();
			Channel channel = (Channel) room.get(name);

			if(channel.push().size() > 0) {
				if(!first) {
					list.append(", ");
				}

				list.append(name + "(" + channel.push().size() + ")");
				first = false;
			}
		}

		return list.toString();
	}

	protected Channel remove(Channel channel) {
		if(channel.push().size() == 0 && !channel.persistent()) {
			return (Channel) room.remove(channel.name());
		}

		return null;
	}

	protected Channel find(String name) {
		return (Channel) room.get(name);
	}

	protected Channel find(Session session) {
		return find(session, "");
	}

	protected Channel find(Session session, String fail) {
		String name = session.string("channel", "lobby");
		boolean persistent = false;

		if(!session.containsKey("channel") && fail.length() > 0) {
			name = fail;
			persistent = true; // Request by HiL
		}

		Channel channel = (Channel) room.get(name);

		if(channel == null) {
			channel = add(name, null, persistent);
		}

		session.put("channel", channel.name());

		return channel;
	}
}
